/*

SNRmax.js: Find image with best SNR
================================================================================

Sucht das Bild mit dem höchstem Signal-Rauschabstand aus einer Auswahl von
Bildern. Der Name des Files mit dem besten Wert kann danach bei der
ImageIntegration als Referenzbild ausgewählt werden.

Searches the image with the highest signal-to-noise ratio from a selection of
images. The name of the file with the best value can then be selected in the
Image Integration as a reference image.
================================================================================

 * * SNRmax
 *
 * Find file with max SNR
 *
 * Input:   fit-files
 *
 * Output:  the filename with maximum SNR found
 *
 * Copyright (C) 2014, Hartmut V. Bornemann
 *
 * Version history
 *          1.0    2014-10-08 first release
*/
#feature-id Image Analysis > SNRmax

#include <pjsr/Color.jsh>
#include <pjsr/FrameStyle.jsh>
#include <pjsr/TextAlign.jsh>
#include <pjsr/StdButton.jsh>
#include <pjsr/Sizer.jsh>

#define StdCursor_ArrowQuestion 17

#define TITLE "SNRmax"
#define VERSION "1.1.0"

#feature-info Find file with max SNR.<br>\
   <br>\Written by Hartmut V. Bornemann.<br>\
   <br>\
   References: None<br>\
   <br>\
#define strict 0

//////////////////////////////////////////////////////////////////////////////
//
// Find file with max SNR Dialog
//
//////////////////////////////////////////////////////////////////////////////


function SNRmaxDialog()
{
   // Add all properties and methods of the core Dialog object to this object.
   this.__base__ = Dialog;
   this.__base__();

   var dlg = this;
   var closing = false;
   var column = 0;
   var maxx;
   var maxFile;
   var h = 24;
   var defaultProcess = "<default>";
   var processIcon = defaultProcess;

   /*
      begin: FileSelection
   */

   var tbx = new TreeBox(this);
   with (tbx)
   {
      //
      // create a treebox with 2 columns
      // column 0 contains filenames
      // column 1 contains the full path
      //
      numberOfColumns = 2;
      position = new Point(8, 8);
      rootDecoration = false;
      height = 240;
      width  = 400;
      multipleSelection = true;
      setHeaderText(0, "Files");
      setHeaderText(1, "Full Path and FileNames");
      hideColumn(1);
      onNodeClicked = function( item, columnIndex )
      {
         enableButtons(false);
      }
   }

   var x = tbx.frameRect.right + 8;
   var y = tbx.frameRect.top;

   var btnAdd = new PushButton(this);
   with (btnAdd)
   {
      text = "Add Files";
      position = new Point(x, y);
      resize( 150 , height);
      y   += height + 8;
      maxx = frameRect.right;
      h    = height - 4;
      onClick  = function()
      {
         var ofd = new OpenFileDialog();
         ofd.multipleSelections = true;
         //ofd.filters = [ [ "FITS or XIFS Files", ".fit", ".fits", ".fts", ".xisf"]];
         ofd.filters = [ [ "FITS Files", ".fit", ".fits"],
                         [ "XIFS Files",  ".xisf"]];

         if (ofd.execute())
         {
            for (var i = 0; i < ofd.fileNames.length; i++)
            {
               var filename =  ofd.fileNames[i];
               //
               // find filename in treebox column 1
               //
               var index = -1;
               for (var i = 0; i < tbx.numberOfChildren; i++)
               {
                  if (tbx.child(i).text(1) == filename)
                  {
                     index = i;
                     break;
                  }
               }
               if (index < 0)
               {
                  var n = new TreeBoxNode(tbx);
                  n.setText(0, File.extractNameAndExtension(filename));
                  n.setText(1, filename);
               }
            }
            enableButtons(false);
            lblCount.text = tbx.numberOfChildren.toString();
         }
      }
   }

   var btnSelAll = new PushButton(this);
   with (btnSelAll)
   {
      enabled = false;
      text = "Select All";
      position = new Point(x, y);
      resize( 150 , height);
      y += height + 8;
      onClick  = function()
      {
         tbx.selectAllNodes();
         enableButtons(false);
      }
   }

   var btnInvertSel = new PushButton(this);
   with (btnInvertSel)
   {
      enabled = false;
      text = "Invert Selection";
      position = new Point(x, y);
      resize( 150 , height);
      y += height + 8;
      onClick  = function()
      {
         for (var i = 0; i < tbx.numberOfChildren; i++)
         {
            var n = tbx.child(i);
            n.selected = !n.selected;
         }
         enableButtons(false);
      }
   }

   var btnRemove = new PushButton(this);
   with (btnRemove)
   {
      enabled = false;
      text = "Remove Selected";
      position = new Point(x, y);
      resize( 150 , height);
      y += height + 8;
      onClick  = function()
      {
         var nodes  = tbx.selectedNodes;
         for (var i = 0; i < nodes.length; i++)
         {
            var c = tbx.childIndex(nodes[i]);
            tbx.remove(c);
         }
         enableButtons(false);
      }
   }

   var btnClear = new PushButton(this);
   with (btnClear)
   {
      enabled = false;
      text = "Clear";
      position = new Point(x, y);
      resize( 150 , height);
      y += height + 8;
      onClick  = function()
      {
         tbx.clear();
         enableButtons(false);
      }
   }

   var cbFullNames = new CheckBox(this);
   with (cbFullNames)
   {
      enabled = false;
      y = tbx.frameRect.bottom - height;
      position = new Point(x, y);
      text = "Full names";
      onCheck = function(checked)
      {
         tbx.hideColumn(column);
         column = 1-column;
         tbx.showColumn(column);
         if (lblFilename.text != "")
         {
            if (checked)
               lblFilename.text = maxFile;
            else
               lblFilename.text = File.extractNameAndExtension(maxFile);
         }
      }
   }

   /*
      end: FileSelection
   */

   x  = tbx.frameRect.left;
   y  = tbx.frameRect.bottom + 16;

   var lblCount = new Label(this);
   with (lblCount)
   {
      backgroundColor = Color.rgbaColor(255, 255, 255, 255);
      position = new Point(x,y);
      resize (80, h);
      textAlignment = TextAlign_Center;
   }

   var btnExec = new PushButton(this);
   with (btnExec)
   {
      enabled = false;
      text = "Execute";
      position = new Point(tbx.frameRect.right - width, y);
      onClick = function()
      {
         enableButtons(true);
         var maxdb = 0;
         var results = new Array();
         var count = tbx.numberOfChildren;
         for (var i = 0; i < tbx.numberOfChildren; i++)
         {
            lblCount.text = format("%d",(i + 1)) + "(" + tbx.numberOfChildren.toString() + ")";
            var input_filename = tbx.child(i).text(1);
            var img = ReadImageFile(input_filename);
            // evaluate mainview channels
            for ( var c = 0; c < img.numberOfChannels; ++c )
            {
               processEvents();
               if (Console.abortRequested || closing)
               {
                  enableButtons(false);
                  Console.warningln("Abort requested");
                  return;
               }
               img.selectedChannel = c;
               var channel = "K";
               if (img.numberOfChannels > 1) channel = "RGB"[c];
               var snr = new SNREvaluation ( img );
               results = results.concat(new frameData(
                  File.extractNameAndExtension(input_filename),
                  snr.db, channel, snr.E.msg));
               if (snr.db > maxdb)
               {
                  maxdb = snr.db;
                  maxFile = input_filename;
                  if (cbFullNames.checked)
                     lblFilename.text = maxFile;
                  else
                     lblFilename.text = File.extractNameAndExtension(maxFile);
                  lblSNR.text = format("%.2f db", maxdb);
               }
            }
            img = null;
         }

         Console.writeln("");
         Console.writeln("Summary:");
         Console.writeln("========");

         results.sort(function(a, b)
            {return a.filename > b.filename});
         for each (var v in results) {
         if (v.channel == "K" | v.channel == "R") Console.writeln("");
         Console.writeln(format("SNR = %.2f db", v.snr) +
            ", " + v.channel + " " + v.filename + " " + v.comment);
         }

         Console.writeln("");
         Console.noteln("Max. SNR found in " + maxFile);
         Console.writeln("End SNRmax");
         Console.flush();
         var info = new MessageBox("Max. SNR found in " + maxFile, "SNRmax");
         info.execute();
         enableButtons(false);
      }
   }

   var btnReady = new PushButton(this);
   with (btnReady)
   {
      text = "Close";
      position = new Point(maxx - width, y);
      onClick = function()
      {
         closing = true;
         dlg.ok();
      }
   }

   // my ©
   x  = tbx.frameRect.left;
   y += 32;

   var lblCopyright = new Label(this)
   with (lblCopyright)
   {
      position  = new Point(8, y);
      text      = "©";
      cursor    = new Cursor(StdCursor_ArrowQuestion);
      onMousePress = function(x, y, button, buttonState, modifiers )
      {
        var dlg = new helpdlg();
        dlg.execute();
      }
   }

   y += 24;

   var lblmaxVal = new Label(this);
   with (lblmaxVal)
   {
      position = new Point(x,y);
      text = "Max. Value:";
   }

   var lblSNR = new Label(this);
   with (lblSNR)
   {
      backgroundColor = Color.rgbaColor(255, 255, 255, 255);
      textAlignment = TextAlign_Center;
      lineWidth = 1;
      resize (80, h);
      position = new Point(lblmaxVal.frameRect.right + 4, y);
   }

   y += 32;

   var lblmaxFile = new Label(this);
   with (lblmaxFile)
   {
      position = new Point(x,y);
      text = "File:";
   }

   var lblFilename = new Label(this);
   with (lblFilename)
   {
      backgroundColor = Color.rgbaColor(255, 255, 255, 255);
      lineWidth = 1;
      textAlignment = TextAlign_Left;
      x = lblmaxVal.frameRect.right + 4;
      resize (btnAdd.frameRect.right - x, h);
      position = new Point(lblmaxVal.frameRect.right + 4, y);
   }

   with (this)
   {
      width  = maxx + 8;
      height = lblmaxFile.frameRect.bottom + 4;
      windowTitle = "Find file with max SNR";
      userResizable = false;
   }

   function enableButtons(executing)
   {
      if (executing)
      {
         btnAdd.enabled    = false;
         btnExec.enabled   = false;
         btnSelAll.enabled = false;
         btnClear.enabled  = false;
         btnRemove.enabled = false;
         btnInvertSel.enabled = false;
         cbFullNames.enabled  = false;
      }
      else
      {
         btnAdd.enabled    = true;
         btnExec.enabled   = tbx.numberOfChildren > 0;
         btnSelAll.enabled = tbx.numberOfChildren > 0;
         btnClear.enabled  = tbx.numberOfChildren > 0;
         btnRemove.enabled = tbx.selectedNodes.length > 0;
         btnInvertSel.enabled = tbx.selectedNodes.length > 0;
         cbFullNames.enabled  = tbx.numberOfChildren > 0;
      }
   }
}

function SNREvaluation ( img )
{
   this.E = new NoiseEvaluation( img );
   var snr = img.meanOfSquares(img.selectedRect)/ this.E.sigma / this.E.sigma;
   this.db = 10 * Math.log10(snr);
}

function frameData(filename, snr, channel, comment)
{
   this.filename = filename;
   this.snr = snr;
   this.channel = channel;
   this.comment = comment;
}

function NoiseEvaluation( img )
{
   this.msg = "";
   var a, n = 4, m = 0.01 * img.selectedRect.area;
   for ( ;; )
   {
      a = img.noiseMRS( n );
      if ( a[1] >= m )
         break;
      if ( --n == 1 )
      {
         this.msg = "k-sigma noise estimate";
         a = img.noiseKSigma();
         break;
      }
   }
   this.sigma = a[0]; // estimated stddev of Gaussian noise
   this.count = a[1]; // number of pixels in the noise pixels set
   this.layers = n;   // number of layers used for noise evaluation
}

function ReadImageFile(filePath)
{
   var suffix = File.extractExtension( filePath ).toLowerCase();
   var F = new FileFormat( suffix, true/*toRead*/, false/*toWrite*/ );
   if ( F.isNull )
      throw new Error( "No installed file format can read \'" + suffix + "\' files." );

   var f = new FileFormatInstance( F );
   if ( f.isNull )
      throw new Error( "Unable to instantiate file format: " + F.name );

   var d = f.open( filePath);
   if ( d.length < 1 )
      throw new Error( "Unable to open file: " + filePath );

   if ( d.length > 1 )
      throw new Error( "Multi-image files are not supported by this script: " + filePath );

   var image = new Image();
   if ( !f.readImage( image ) )
      throw new Error( "Unable to read file: " + filePath );
   f.close();
   return image;
}

SNRmaxDialog.prototype = new Dialog;


function helpdlg()
{
   this.__base__ = Dialog;
   this.__base__();

   this.minHeight = 400;
   this.minWidth  = 500;

   this.windowTitle = "SNRmax Dialog Help";

   this.OK_Button = new PushButton (this);
   with (this.OK_Button)
   {
      text = "OK";
      maxWidth = 100;

      onClick = function()
      {
         this.parent.ok();
      }
   }

   this.helpText = new TextBox( this );
   with ( this.helpText )
   {
      frameStyle = FrameStyle_Box;
      margin = 4;
      readOnly = true;
      wordWrapping = true;
      useRichText = true;

      text = "<b>" + TITLE + " V" + VERSION +
      " Find file with the maximum SNR.</b>" +
      "<p>Read and evaluate a list of image files and find the best one with the highest SNR (Signal to Noise Ratio). " +
      " This can be helpful to know the reference image for the " +
      "<b>ImageIntegration</b> process.</p>" +
      "Usage: select files and press 'Execute'. The filename with the best SNR is shown in the bottom label of the dialog window at the end of successful execution." +
      "<p>Early termination is possible by clicking Pause/Abort in the console window.</p>" +
      "<p>Copyright © Hartmut Bornemann, 2014, mailto:hvb356@yahoo.com</p>";
   }

   this.buttonSizer = new HorizontalSizer;
   with ( this.buttonSizer )
   {
      spacing = 4;
      add (this.OK_Button);
   }


   this.sizer = new VerticalSizer;
   with ( this.sizer )
   {
      margin = 6;
      spacing = 6;
      add( this.helpText );
      add( this.buttonSizer );
   }

   this.helpText.home();
}

helpdlg.prototype = new Dialog;

function main()
{
   Console.show();
   Console.abortEnabled = true;
   var dialog = new SNRmaxDialog();
   dialog.execute();
}

main();

// ****************************************************************************
// EOF SNRmax.js - Released 2014/10/08 12:00:00 UTC
